import eagerx
import numpy as np
from eagerx.core.specs import ProcessorSpec


class MirrorAction(eagerx.Processor):
    @classmethod
    def make(cls, index=0, constant=0, offset=0, scale=1) -> ProcessorSpec:
        spec = cls.get_specification()
        spec.config.update(offset=offset, index=index, scale=scale, constant=constant)
        return spec

    def initialize(self, spec: ProcessorSpec):
        self.offset = spec.config.offset
        self.index = spec.config.index
        self.scale = spec.config.scale
        self.constant = spec.config.constant

    def convert(self, msg):
        action = self.scale * msg.data[self.index] + self.constant
        mirrored = np.array([action, -action], dtype="float32")
        mirrored += self.offset
        return mirrored



import os
import eagerx_franka
import yaml
from eagerx_utility.utils import launch_node


def get_joint_limits(robot_model: str, joint_limits: str):

    config_path = os.path.dirname(eagerx_franka.__file__) + f"/../assets/robots/{robot_model}"
    try:
        joint_limits = joint_limits if isinstance(joint_limits, str) else f"{config_path}/joint_limits.yaml"
        with open(joint_limits, "r") as yamlfile:
            joint_limits = yaml.safe_load(yamlfile)
    except IOError:
        print(f"Joint Limits File was not found in: {config_path}")
        raise
    return joint_limits


def generate_urdf(
    robot: str,
    ns="",
    use_gripper=True,
):
    import rospy
    import rosparam

    module_path = os.path.dirname(eagerx_franka.__file__) + "/../assets/"
    launch_file = module_path + "franka_description.launch"
    cli_args = [
        f"module_path:={module_path}",
        f"robot:={robot}",
        f"ns:={ns}",
        f"use_gripper:={use_gripper}",
    ]
    launch = launch_node(launch_file, cli_args)
    launch.start()

    # Replace mesh urls
    urdf_key = f"{ns}/{robot}/robot_description"
    urdf: str = rospy.get_param(urdf_key)
    urdf_sbtd = urdf.replace("package://franka_description/", module_path)
    rosparam.upload_params(urdf_key, urdf_sbtd)

    return urdf_key



import eagerx
from eagerx import Space
from eagerx_pybullet.engine import PybulletEngine
from eagerx.core.specs import ObjectSpec
from eagerx.core.graph_engine import EngineGraph
import eagerx.core.register as register
import eagerx_franka
from eagerx_franka.utils import generate_urdf, get_joint_limits

import os

try:
    from urdf_parser_py.urdf import URDF
except ImportError:
    print("Please install urdf_parser_py")


class FrankaArm(eagerx.Object):
    @classmethod
    @register.sensors(
        position=Space(dtype="float32"),
        velocity=Space(dtype="float32"),
        force_torque=Space(low=-20, high=20, shape=(6,), dtype="float32"),
        ee_pos=Space(low=[-2, -2, 0], high=[2, 2, 2], dtype="float32"),
        ee_orn=Space(low=-1, high=1, shape=(4,), dtype="float32"),
    )
    @register.actuators(
        pos_control=Space(dtype="float32"),
        vel_control=Space(dtype="float32"),
        gripper_control=Space(low=[0], high=[1], dtype="float32"),
    )
    @register.engine_states(
        position=Space(dtype="float32"),
        velocity=Space(dtype="float32"),
        gripper=Space(low=[0.5], high=[0.5], dtype="float32"),
    )
    def make(
        cls,
        name: str,
        robot_type: str,
        arm_name=None,
        sensors=None,
        actuators=None,
        states=None,
        rate=30,
        base_pos=None,
        base_or=None,
        self_collision=False,
        fixed_base=True,
        joint_limits=None,
        regenerate_urdf=False,
    ) -> ObjectSpec:
        """Object spec of FrankaArm"""
        if URDF is None:
            raise ImportError("Ros not installed. Required for generating urdf.")

        spec = cls.get_specification()
        joint_limits = get_joint_limits(robot_type, joint_limits)

        # Extract info on franka arm from assets
        if regenerate_urdf:
            urdf = URDF.from_parameter_server(generate_urdf(robot_type, ns="pybullet_urdf"))
        else:
            module_path = os.path.dirname(eagerx_franka.__file__) + "/../assets/"
            urdf = URDF.from_xml_file(module_path + "franka_panda/panda.urdf")

        gripper_names = [joint.name for joint in urdf.joints if "finger" in joint.name]
        gripper_link = [link.name for link in urdf.links if f"{robot_type}_link8" in link.name][0]

        # Determine joint limits
        joint_names = []
        joint_lower, joint_upper, vel_limit = [], [], []
        for n in joint_limits.keys():
            joint_names.append(f"{robot_type}_{n}")
            joint_obj = next((joint for joint in urdf.joints if joint.name == f"{robot_type}_{n}"), None)
            joint_lower.append(joint_obj.limit.lower)
            joint_upper.append(joint_obj.limit.upper)
            vel_limit.append(joint_obj.limit.velocity)

        # Determine gripper limits
        gripper_lower, gripper_upper = [], []
        for n in gripper_names:
            joint_obj = next((joint for joint in urdf.joints if joint.name == n), None)
            gripper_lower.append(joint_obj.limit.lower)
            gripper_upper.append(joint_obj.limit.upper)

        # Modify default config
        spec.config.name = name
        spec.config.sensors = sensors if isinstance(sensors, list) else ["position", "velocity"]
        spec.config.actuators = actuators if isinstance(actuators, list) else ["pos_control", "gripper_control"]
        spec.config.states = states if isinstance(states, list) else ["position", "velocity", "gripper"]

        # Add registered config params
        spec.config.robot_type = robot_type
        spec.config.arm_name = arm_name if arm_name else robot_type
        spec.config.joint_names = joint_names
        spec.config.gripper_names = gripper_names
        spec.config.gripper_link = gripper_link
        spec.config.gripper_lower = gripper_lower
        spec.config.gripper_upper = gripper_upper
        spec.config.base_pos = base_pos if base_pos else [0, 0, 0]
        spec.config.base_or = base_or if base_or else [0, 0, 0, 1]
        spec.config.self_collision = self_collision
        spec.config.fixed_base = fixed_base
        spec.config.joint_limits = joint_limits
        spec.config.joint_lower = joint_lower
        spec.config.joint_upper = joint_upper
        spec.config.vel_limit = vel_limit
        spec.config.urdf = urdf.to_xml_string()
        spec.config.regenerate_urdf = regenerate_urdf

        # Set rates
        spec.sensors.position.rate = rate
        spec.sensors.velocity.rate = rate
        spec.sensors.force_torque.rate = rate
        spec.actuators.pos_control.rate = rate
        spec.actuators.vel_control.rate = rate
        spec.actuators.gripper_control.rate = 1

        # Set variable spaces
        spec.sensors.position.space.update(low=joint_lower, high=joint_upper)
        spec.sensors.velocity.space.update(low=[-v for v in vel_limit], high=vel_limit)
        spec.sensors.ee_pos.rate = rate
        spec.sensors.ee_orn.rate = rate
        spec.actuators.pos_control.space.update(low=joint_lower, high=joint_upper)
        spec.actuators.vel_control.space.update(low=[-v for v in vel_limit], high=vel_limit)
        spec.states.position.space.update(low=[0.0 for _j in joint_lower], high=[0.0 for _j in joint_upper])
        spec.states.velocity.space.update(low=[0.0 for _j in joint_lower], high=[0.0 for _j in joint_upper])
        return spec

    @staticmethod
    @register.engine(PybulletEngine)
    def pybullet_engine(spec: ObjectSpec, graph: EngineGraph):
        import pybullet as pb

        """Engine-specific implementation (Pybullet) of the object."""
        # Set object arguments (as registered per register.engine_params(..) above the engine.add_object(...) method.)

        module_path = os.path.dirname(eagerx_franka.__file__) + "/../assets/franka_panda/"
        urdf = spec.config.urdf
        if not spec.config.regenerate_urdf:
            urdf_sbtd = urdf.replace("package://", module_path)
            spec.engine.urdf = urdf_sbtd
        else:
            spec.engine.urdf = urdf
        spec.engine.basePosition = spec.config.base_pos
        spec.engine.baseOrientation = spec.config.base_or
        spec.engine.fixed_base = spec.config.fixed_base
        spec.engine.self_collision = spec.config.self_collision
        spec.engine.flags = pb.URDF_ENABLE_CACHED_GRAPHICS_SHAPES

        lower = spec.config.gripper_lower
        upper = spec.config.gripper_upper
        constant = abs(float(lower[0]))
        scale = float(upper[0]) - float(lower[0])

        # Create engine_states (no agnostic states defined in this case)
        from eagerx_franka.franka_arm.pybullet.enginestates import FrankaGripper
        from eagerx_pybullet.enginestates import JointState

        joints = spec.config.joint_names
        spec.engine.states.gripper = FrankaGripper.make(spec.config.gripper_names, constant, scale)
        spec.engine.states.position = JointState.make(joints=joints, mode="position")
        spec.engine.states.velocity = JointState.make(joints=joints, mode="velocity")

        # Create sensor engine nodes
        from eagerx_pybullet.enginenodes import LinkSensor, JointSensor, JointController

        pos_sensor = JointSensor.make("pos_sensor", rate=spec.sensors.position.rate, process=2, joints=joints, mode="position")
        vel_sensor = JointSensor.make("vel_sensor", rate=spec.sensors.velocity.rate, process=2, joints=joints, mode="velocity")
        ft_sensor = JointSensor.make(
            "ft_sensor",
            rate=spec.sensors.force_torque.rate,
            process=2,
            joints=[spec.config.joint_names[-1]],
            mode="force_torque",
        )
        ee_pos_sensor = LinkSensor.make(
            "ee_pos_sensor",
            rate=spec.sensors.ee_pos.rate,
            links=[spec.config.gripper_link],
            mode="position",
        )
        ee_orn_sensor = LinkSensor.make(
            "ee_orn_sensor",
            rate=spec.sensors.ee_orn.rate,
            links=[spec.config.gripper_link],
            mode="orientation",
        )

        # Create actuator engine nodes
        # Rate=None, but we will connect it to an actuator (thus will use the rate set in the agnostic specification)
        pos_control = JointController.make(
            "pos_control",
            rate=spec.actuators.pos_control.rate,
            joints=joints,
            mode="position_control",
            vel_target=len(joints) * [0.0],
            pos_gain=len(joints) * [0.5],
            vel_gain=len(joints) * [1.0],
            max_vel=[0.5 * vel for vel in spec.config.vel_limit],
            max_force=len(joints) * [10.0],
        )
        vel_control = JointController.make(
            "vel_control",
            rate=spec.actuators.vel_control.rate,
            joints=joints,
            mode="velocity_control",
            vel_gain=len(joints) * [1.0],
            max_force=len(joints) * [25.0],  # todo: limit?
        )

        gripper = JointController.make(
            "gripper_control",
            rate=spec.actuators.gripper_control.rate,
            joints=spec.config.gripper_names,
            mode="position_control",
            vel_target=[0.0, 0.0],
            pos_gain=[0.5, 0.5],
            vel_gain=[1.0, 1.0],
            max_force=[2.0, 2.0],
        )
        from eagerx_franka.franka_arm.processor import MirrorAction

        gripper.inputs.action.processor = MirrorAction.make(index=0, constant=constant, scale=scale)

        # Connect all engine nodes
        graph.add([pos_sensor, vel_sensor, ft_sensor, ee_pos_sensor, ee_orn_sensor, pos_control, vel_control, gripper])
        graph.connect(source=pos_sensor.outputs.obs, sensor="position")
        graph.connect(source=vel_sensor.outputs.obs, sensor="velocity")
        graph.connect(source=ft_sensor.outputs.obs, sensor="force_torque")
        graph.connect(source=ee_pos_sensor.outputs.obs, sensor="ee_pos")
        graph.connect(source=ee_orn_sensor.outputs.obs, sensor="ee_orn")
        graph.connect(actuator="pos_control", target=pos_control.inputs.action)
        graph.connect(actuator="vel_control", target=vel_control.inputs.action)
        graph.connect(actuator="gripper_control", target=gripper.inputs.action)



from scipy.spatial.transform import Rotation as R
import typing as t
import eagerx
import numpy as np
import gym


# Define environment
class ArmEnv(eagerx.BaseEnv):
    def __init__(self, name, rate, graph, engine, backend, max_steps: int, add_bias: bool = False, exclude_z: bool = True):
        super().__init__(name, rate, graph, engine, backend=backend, force_start=False)
        self.steps = 0
        self.max_steps = max_steps

        # Exclude
        self._exclude_z = exclude_z
        self._exclude_list = ["pos", "pos_desired"]

        # Bias
        self._add_bias = add_bias
        self.solid_bias = None
        self.yaw_bias = None

        # Rwd publishers
        self._pub_rwd = self.backend.Publisher(f"{self.ns}/environment/reward", "float32")

    @property
    def observation_space(self) -> gym.spaces.Space:
        obs_space = self._observation_space
        if self._exclude_z:
            names = obs_space.keys()
            for o in self._exclude_list:
                if o in names:
                    low, high = obs_space[o].low[:, :2], obs_space[o].high[:, :2]
                    obs_space[o] = gym.spaces.Box(low=low, high=high, dtype="float32")
        return obs_space

    def _exclude_obs(self, obs):
        names = obs.keys()
        for o in self._exclude_list:
            if o in names:
                obs[o] = obs[o][:, :2]
        return obs

    def step(self, action):
        # Step the environment
        self.steps += 1
        info = dict()
        obs = self._step(action)

        # Calculate reward
        yaw = obs["yaw"][0]
        yaw_des = obs["yaw_desired"][0]
        force = obs["force_torque"][0] if len(obs["force_torque"][0]) > 0 else 3 * [0.0]
        ee_pos = obs["ee_position"][0]
        goal_pos = obs["pos_desired"][0]
        achieved_pos = obs["pos"][0]
        vel = obs["velocity"][0] if "velocity" in obs else np.array([0], dtype="float32")
        des_vel = action["velocity"][0] if "velocity" in action else 0 * vel
        # Penalize distance of the end-effector to the object
        rwd_near = 0.4 * -abs(np.linalg.norm(ee_pos - obs["pos"][0]) - 0.05)
        # Penalize distance of the object to the goal
        yaw_error = min(abs(yaw_des - yaw), abs(yaw_des - 0.5 * np.pi - yaw), abs(yaw_des + 0.5 * np.pi - yaw))
        pos_error = np.linalg.norm(goal_pos - achieved_pos)
        rwd_pos = -4.0 * pos_error
        rwd_or = -((yaw_error / (1.0 + 10.0 * pos_error)) ** 2)
        # Penalize actions (indirectly, by punishing the angular velocity.
        rwd_ctrl = 0.1 * -np.linalg.norm(des_vel - vel)
        # Penalize force applied to box in vertical direction
        rwd_force = -0.0001 * (force[1] - 3.2) ** 2
        rwd = rwd_pos + rwd_ctrl + rwd_near + rwd_force + rwd_or
        # Print rwd build-up
        out_of_reach = np.linalg.norm(achieved_pos[:2]) > 1.0
        if out_of_reach:
            rwd = -50

        timed_out = self.steps >= self.max_steps
        if timed_out:
            info["TimeLimit.truncated"] = True

        done = out_of_reach | timed_out

        # Simulate bias in observations.
        if self._add_bias:
            obs["pos"] += self.solid_bias
            obs["yaw"] = (obs["yaw"] + self.yaw_bias) % (np.pi / 2)

        # Exclude z observations
        if self._exclude_z:
            obs = self._exclude_obs(obs)

        # Publish reward
        self._pub_rwd.publish(np.array([rwd, rwd_pos, rwd_or, rwd_ctrl, rwd_force, rwd_near], dtype="float32"))

        return obs, rwd, done, info

    def reset(self, states: t.Optional[t.Dict[str, np.ndarray]] = None):
        # Reset steps counter
        self.steps = 0

        # Sample bias
        self.solid_bias = 0.5 * 0.02 * (2 * np.random.random((3,)) - 1)
        self.yaw_bias = 0.5 * (0.1 * np.pi / 2) * (2 * np.random.random() - 1)

        # Sample states
        _states = self.state_space.sample()

        # Sample new starting orientation (vary yaw)
        yaw = np.random.random(()) * np.pi / 2
        self.state_space["solid/orientation"] = R.from_euler("zyx", [yaw, 0.0, 0.0]).as_quat().astype("float32")

        # Sample new starting state (at least 17 cm from goal)
        radius = 0.17
        while True:
            solid_pos = self.state_space["solid/position"].sample()
            goal_pos = self.state_space["goal/position"].sample()
            if np.linalg.norm(solid_pos[:2] - goal_pos[:2]) > radius:
                _states["solid/position"] = solid_pos
                _states["goal/position"] = goal_pos
                break

        # Set initial position state
        if "solid/aruco/position" in _states:
            self.state_space["solid/aruco/position"] = solid_pos

        # Overwrite with user provided states
        if states is not None:
            for key, value in states:
                if key in _states and value.shape == _states[key].shape:
                    _states[key] = value
                else:
                    self.backend.logwarn(f"State `{key}` incorrectly specified.")

        # Perform reset
        obs = self._reset(_states)

        # Simulate bias in observations.
        if self._add_bias:
            obs["pos"] += self.solid_bias
            obs["yaw"] = (obs["yaw"] + self.yaw_bias) % (np.pi / 2)

        # Exclude z observations
        if self._exclude_z:
            obs = self._exclude_obs(obs)
        return obs



import typing as t
import numpy as np
import eagerx
from eagerx import Space
from eagerx.core.specs import NodeSpec
import eagerx.core.register as register
from eagerx.utils.utils import Msg
import roboticstoolbox as rtb
from spatialmath import SE3
from scipy.spatial.transform import Rotation as R
import modern_robotics as mr


class EndEffectorDownward(eagerx.Node):
    @classmethod
    def make(
        cls,
        name: str,
        rate: float,
        Slist: t.List[t.List[float]],
        M: t.List[t.List[float]],
        upper: t.List[float],
        lower: t.List[float],
        max_dxyz: t.List[float],
        max_dyaw: float,
        min_z: float = 0.03,
        eomg: float = 0.001,
        ev: float = 0.001,
        process: int = eagerx.NEW_PROCESS,
    ) -> NodeSpec:
        """
        Calculate desired joint positions based on delta pose of the end effector in task space using inverse kinematics.

        :param name: Node name.
        :param rate: Rate at which callback is called.
        :param Slist: The joint screw axes in the space frame when the manipulator is at the home position,
                      in the format of a matrix with axes as the columns. See modern robotics toolkit.
        :param M: The home configuration of the end-effector.
        :param upper: Upper joint limits.
        :param lower: Lower joint limits.
        :param max_dxyz: Maximum delta position in xyz [m/s].
        :param max_dyaw: Maximum delta rotation in yaw [rad/s].
        :param min_z: Minimum z position of the end-effector's link (cog?) [m].
        :param eomg: A small positive tolerance on the end-effector orientation
                     error. The returned joint angles must give an end-effector
                     orientation error less than eomg. See modern robotics toolkit.
        :param ev: A small positive tolerance on the end-effector linear position
                   error. The returned joint angles must give an end-effector
                   position error less than ev. See modern robotics toolkit.
        :param process: {0: NEW_PROCESS, 1: ENVIRONMENT, 2: ENGINE, 3: EXTERNAL}.
        :return: Parameter specification of the node.
        """
        spec = cls.get_specification()

        # Modify default node params
        spec.config.name = name
        spec.config.rate = rate
        spec.config.process = process
        spec.config.inputs = ["dxyz", "dyaw", "current", "xyz", "orn"]
        spec.config.outputs = ["target", "dtarget"]

        # Modify custom node params
        spec.config.Slist = Slist
        spec.config.M = M
        spec.config.max_dxyz = max_dxyz
        spec.config.max_dyaw = max_dyaw
        spec.config.eomg = eomg
        spec.config.ev = ev
        spec.config.min_z = min_z

        # Add converter & space
        spec.inputs.dxyz.space.update(low=[-i for i in max_dxyz], high=max_dxyz)
        spec.inputs.dyaw.space.update(low=-max_dyaw, high=max_dyaw)
        spec.inputs.current.space.update(low=lower, high=upper)
        return spec

    def initialize(self, spec: NodeSpec):
        np.set_printoptions(precision=2, suppress=True)
        self.Slist = np.array(spec.config.Slist, dtype="float32")
        self.M = np.array(spec.config.M, dtype="float32")
        self.max_dxyz = np.array(spec.config.max_dxyz, dtype="float32")
        self.max_dyaw = np.array(spec.config.max_dyaw, dtype="float32")
        self.eomg = spec.config.eomg
        self.ev = spec.config.ev
        self.min_z = spec.config.min_z

    @register.states()
    def reset(self):
        pass

    @register.inputs(
        dxyz=Space(shape=(3,), dtype="float32"),
        xyz=Space(shape=(3,), dtype="float32"),
        orn=Space(shape=(4,), dtype="float32"),
        dyaw=Space(shape=(), dtype="float32"),
        current=Space(dtype="float32"),
    )
    @register.outputs(target=Space(dtype="float32"), dtarget=Space(dtype="float32"))
    def callback(self, t_n: float, dxyz: Msg, xyz: Msg, orn: Msg, dyaw: Msg, current: Msg):
        dxyz = dxyz.msgs[-1]
        xyz = xyz.msgs[-1]
        orn = orn.msgs[-1]
        dyaw = dyaw.msgs[-1]
        current = current.msgs[-1]

        # Limit dz
        dz = dxyz[-1]
        z = xyz[-1]
        dxyz[-1] = max(dz, self.max_dxyz[-1] * (-1 + 1 / np.exp(max(0, 10 * (z - self.min_z)))))

        # Calculate the target pose
        rot_ee2b = R.from_quat(orn).as_matrix()
        yaw = np.arctan2(rot_ee2b[1, 0], rot_ee2b[0, 0])
        yaw_target = (yaw + dyaw / self.rate) % (2 * np.pi)
        yaw_target_cos = np.cos(yaw_target)
        yaw_target_sin = np.sin(yaw_target)
        rot_t2b = np.array([[yaw_target_cos, yaw_target_sin, 0], [yaw_target_sin, -yaw_target_cos, 0], [0, 0, -1]])

        # 4x4 Transformation Matrix representing the transform from the
        # /<robot_name>/base_link frame to the /<robot_name>/ee_gripper_link frame
        T_sd = np.identity(4)
        T_sd[:3, :3] = rot_t2b
        T_sd[:3, 3] = xyz + dxyz / self.rate  # Scale delta position with rate

        theta_list, success = mr.IKinSpace(self.Slist, self.M, T_sd, current, self.eomg, self.ev)
        if success:
            target = np.array(theta_list, dtype="float32")
            dtarget = (target - current) * self.rate  # [rad/sec]
        else:
            # self.backend.logwarn("no solution")
            target = current
            dtarget = current * 0
        return dict(target=target, dtarget=dtarget)


class PandaEndEffectorDownward(eagerx.Node):
    @classmethod
    def make(
        cls,
        name: str,
        rate: float,
        max_dxyz: t.List[float],
        max_dyaw: float,
        min_z: float = 0.03,
        process: int = eagerx.NEW_PROCESS,
    ) -> NodeSpec:
        """
        Calculate desired joint positions based on delta pose of the end effector in task space using inverse kinematics.

        :param name: Node name.
        :param rate: Rate at which callback is called.
        :param max_dxyz: Maximum delta position in xyz [m/s].
        :param max_dyaw: Maximum delta rotation in yaw [rad/s].
        :param min_z: Minimum z position of the end-effector's link (cog?) [m].
        :param process: {0: NEW_PROCESS, 1: ENVIRONMENT, 2: ENGINE, 3: EXTERNAL}.
        :return: Parameter specification of the node.
        """
        spec = cls.get_specification()

        # Modify default node params
        spec.config.name = name
        spec.config.rate = rate
        spec.config.process = process
        spec.config.inputs = ["dxyz", "dyaw", "current", "xyz", "orn"]
        spec.config.outputs = ["target", "dtarget"]

        # Modify custom node params
        spec.config.max_dxyz = max_dxyz
        spec.config.max_dyaw = max_dyaw
        spec.config.min_z = min_z

        # Add converter & space
        spec.inputs.dxyz.space.update(low=[-i for i in max_dxyz], high=max_dxyz)
        spec.inputs.dyaw.space.update(low=-max_dyaw, high=max_dyaw)
        return spec

    def initialize(self, spec: NodeSpec):
        np.set_printoptions(precision=2, suppress=True)
        self.max_dxyz = np.array(spec.config.max_dxyz, dtype="float32")
        self.max_dyaw = np.array(spec.config.max_dyaw, dtype="float32")
        self.min_z = spec.config.min_z
        self.robot = rtb.models.Panda()

    @register.states()
    def reset(self):
        pass

    @register.inputs(
        dxyz=Space(shape=(3,), dtype="float32"),
        xyz=Space(shape=(3,), dtype="float32"),
        orn=Space(shape=(4,), dtype="float32"),
        dyaw=Space(shape=(), dtype="float32"),
        current=Space(dtype="float32"),
    )
    @register.outputs(target=Space(dtype="float32"), dtarget=Space(dtype="float32"))
    def callback(self, t_n: float, dxyz: Msg, xyz: Msg, orn: Msg, dyaw: Msg, current: Msg):
        dxyz = dxyz.msgs[-1]
        xyz = xyz.msgs[-1]
        orn = orn.msgs[-1]
        dyaw = dyaw.msgs[-1]
        current = current.msgs[-1]

        # Limit dz
        dz = dxyz[-1]
        z = xyz[-1]
        dxyz[-1] = max(dz, self.max_dxyz[-1] * (-1 + 1 / np.exp(max(0, 10 * (z - self.min_z)))))

        # Calculate the target pose
        rot_ee2b = R.from_quat(orn).as_matrix()
        yaw = np.arctan2(rot_ee2b[0, 1], rot_ee2b[0, 0])
        yaw_target = (yaw + dyaw / self.rate) % (2 * np.pi)

        trans = xyz + dxyz / self.rate  # Scale delta position with rate

        tep = SE3.Trans(trans) * SE3.RPY(yaw_target, np.pi, 0, order="zyx")
        sol = self.robot.ik_lm_chan(tep)
        target = np.asarray(sol[0], dtype="float32")
        dtarget = (target - current) * self.rate  # [rad/sec]
        return dict(target=target, dtarget=dtarget)



import gym
from gym import spaces
import numpy as np
from eagerx import BaseEnv


class GoalArmEnv(gym.Wrapper):
    def __init__(self, env: BaseEnv, add_bias: bool = False):
        super().__init__(env)
        self._add_bias = add_bias
        self._env = env
        flattened_space = spaces.Dict()
        space = dict(self._env.observation_space)
        space.pop("force_torque")
        self._og_des_pos_space = space.pop("pos_desired")
        flat_des_pos_space = spaces.flatten_space(self._og_des_pos_space)
        self._og_des_yaw_space = space.pop("yaw_desired")
        flat_des_yaw_space = spaces.flatten_space(self._og_des_yaw_space)
        self._og_desired_goal_space = spaces.Box(
            low=np.concatenate([flat_des_pos_space.low, flat_des_yaw_space.low]),
            high=np.concatenate([flat_des_pos_space.high, flat_des_yaw_space.high]),
        )

        self._og_observation_space = spaces.Dict(space)

        self._og_achieved_pos_space = space["pos"]
        flat_achieved_pos_space = spaces.flatten_space(self._og_achieved_pos_space)
        self._og_achieved_yaw_space = space["yaw"]
        flat_achieved_yaw_space = spaces.flatten_space(self._og_achieved_yaw_space)
        self._og_achieved_goal_space = spaces.Box(
            low=np.concatenate([flat_achieved_pos_space.low, flat_achieved_yaw_space.low]),
            high=np.concatenate([flat_achieved_pos_space.high, flat_achieved_yaw_space.high]),
        )

        flattened_space["desired_goal"] = spaces.flatten_space(self._og_desired_goal_space)
        flattened_space["observation"] = spaces.flatten_space(self._og_observation_space)
        flattened_space["achieved_goal"] = spaces.flatten_space(self._og_achieved_goal_space)
        self.observation_space = flattened_space
        self.action_space = spaces.flatten_space(self._env.action_space)

    def action(self, action: object) -> object:
        return spaces.unflatten(self._env.action_space, action)

    def observation(self, obs):
        goal_obs = dict()
        obs.pop("force_torque")
        flat_des_pos = spaces.flatten(self._og_des_pos_space, obs.pop("pos_desired"))
        flat_des_yaw = spaces.flatten(self._og_des_yaw_space, obs.pop("yaw_desired"))

        goal_obs["desired_goal"] = spaces.flatten(self._og_desired_goal_space, np.concatenate([flat_des_pos, flat_des_yaw]))

        flat_achieved_pos = spaces.flatten(self._og_achieved_pos_space, obs["pos"])
        flat_achieved_yaw = spaces.flatten(self._og_achieved_yaw_space, obs["yaw"])

        goal_obs["achieved_goal"] = spaces.flatten(
            self._og_achieved_goal_space, np.concatenate([flat_achieved_pos, flat_achieved_yaw])
        )
        # Simulate bias in observations.
        if self._add_bias:
            obs["pos"] = obs["pos"] + self.solid_bias
            obs["yaw"] = (obs["yaw"] + self.yaw_bias) % (np.pi / 2)

        goal_obs["observation"] = spaces.flatten(self._og_observation_space, obs)
        return goal_obs

    def step(self, action):
        action = dict(self.action(action))
        obs, rwd, done, info = self._env.step(action)

        # Calculate rewards independent of goal for HER
        force = obs["force_torque"][0] if len(obs["force_torque"][0]) > 0 else 3 * [0.0]
        ee_pos = obs["ee_position"][0]
        pos = obs["pos"][0]
        vel = obs["velocity"][0] if "velocity" in obs else np.array([0], dtype="float32")
        des_vel = action["velocity"][0] if "velocity" in action else 0 * vel
        rwd_near = 0.4 * -abs(np.linalg.norm(ee_pos - pos) - 0.05)
        rwd_ctrl = 0.1 * -np.linalg.norm(des_vel - vel)
        rwd_force = -0.0001 * (force[1] - 3.2) ** 2
        info["rwd_goal_independent"] = rwd_near + rwd_ctrl + rwd_force

        return self.observation(obs), rwd, done, info

    def reset(self):
        obs = self._env.reset()
        return self.observation(obs)

    def compute_reward(self, achieved_goal, desired_goal, info):
        pos_desired = desired_goal[:, :2]
        yaw_desired = desired_goal[:, -1]
        pos = achieved_goal[:, :2]
        yaw = achieved_goal[:, -1]

        # Penalize distance of the object to the pos_desired
        yaw_error = np.min(
            np.vstack(
                [np.abs(yaw_desired - yaw), np.abs(yaw_desired - 0.5 * np.pi - yaw), np.abs(yaw_desired + 0.5 * np.pi - yaw)]
            ),
            axis=0,
        )
        dist = pos_desired - pos
        pos_error = np.linalg.norm(dist.reshape(-1, 2), axis=1)
        rwd_dist = -4.0 * pos_error - (yaw_error / (1 + 10 * pos_error)) ** 2
        # Add goal independent rewards
        rwd_goal_independent = np.asarray([i["rwd_goal_independent"] for i in info])
        rwd = rwd_dist + rwd_goal_independent
        out_of_reach = np.linalg.norm(achieved_goal[:, :2], axis=1) > 1.0
        rwd[out_of_reach] = -50
        return rwd



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['eagerx_franka',
 'eagerx_franka.franka_arm',
 'eagerx_franka.franka_arm.pybullet',
 'eagerx_franka.ik',
 'eagerx_franka.solid',
 'eagerx_franka.solid.assets']

package_data = \
{'': ['*']}

install_requires = \
['eagerx-pybullet>=0.1.11,<0.2.0',
 'eagerx-utility>=0.0.2,<0.0.3',
 'eagerx>=0.1.33,<0.2.0',
 'modern-robotics>=1.1.0,<2.0.0',
 'moviepy>=1.0.3,<2.0.0',
 'numpy<1.20.0',
 'roboticstoolbox-python>=1.0.3,<2.0.0',
 'scipy>=1.9.3,<2.0.0',
 'stable-baselines3>=1.6.2,<2.0.0',
 'tensorboard>=2.11.0,<3.0.0',
 'urdf-parser-py>=0.0.4,<0.0.5']

setup_kwargs = {
    'name': 'eagerx-franka',
    'version': '0.0.2',
    'description': 'EAGERx interface to franka robot arms.',
    'long_description': 'None',
    'author': 'Jelle Luijkx',
    'author_email': 'j.d.luijkx@tudelft.nl',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/eager-dev/eagerx_franka',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
